from typing import Dict
from sqlalchemy import  Integer, String, Float, ForeignKey, Boolean
from sqlalchemy.orm import relationship, Mapped, mapped_column
from database_service import Base

# Models
class Condition(Base):
    """
    Model for conditions, applicable to creatures

    Source: DND5e(2014) PHB, p.290-292
    """
    __tablename__ = "conditions"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(20), nullable=False, primary_key=True)
    desc: Mapped[str] = mapped_column(String, nullable=False)


class Alignment(Base):
    """
    Alignments
    """
    __tablename__ = "alignments"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(50), nullable=False, primary_key=True)    
    abbreviation: Mapped[str] = mapped_column(String(50), nullable=False, primary_key=True)    
    desc: Mapped[str] = mapped_column(String, nullable=False)


class Monster(Base):
    """
    Model for defining a monster, official or custom
    Source: DND5e(2014) MM
    """
    __tablename__ = "monsters"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(50), nullable=False, primary_key=True)
    desc: Mapped[str] = mapped_column(String, nullable=True)

    type_id: Mapped[int] = mapped_column(Integer, ForeignKey('creature_types.id'))
    type: Mapped["CreatureType"] = relationship(
        "CreatureType",
    )
    
    sub_type_id: Mapped[int] = mapped_column(Integer, ForeignKey('creature_sub_types.id'))
    sub_type: Mapped["CreatureSubType"] = relationship(
        "CreatureSubType",
    )

    size_id: Mapped[int] = mapped_column(Integer, ForeignKey('creature_sizes.id'))
    size: Mapped["CreatureSize"] = relationship(
        "CreatureSize",
    )

    alignment: Mapped[int] = mapped_column(Integer, ForeignKey('alignments.id'))
    alignment: Mapped["Alignment"] = relationship(
        "Alignment",
    )


    armor_class: Mapped[int] = mapped_column(Integer)
    armors: Mapped["MonsterArmor"] = relationship('MonsterArmor', back_populates="monster") 
    speed: Mapped[str]

    # ability scores
    strength: Mapped[int] = mapped_column(Integer)
    # storage is cheap, computation is expensive, store this
    strength_mod: Mapped[int] = mapped_column(Integer)

    dexterity: Mapped[int] = mapped_column(Integer)
    dexterity_mod: Mapped[int] = mapped_column(Integer)

    # will frequently be filtered by these
    challenge_rating: Mapped[Float] = mapped_column(Float, index=True)
    xp_rating: Mapped[int] = mapped_column(Integer, index=True)

    hit_dice: Mapped[str] = mapped_column(String)
    hit_point_formula: Mapped[str] = mapped_column(String)
    hit_point_average: Mapped[int] = mapped_column(Integer)

    # NOTE: if need actual combat automation, make the following fields beloe as junction tables
    # TODO: make this into junction table, so we can take list and then modify and format each element
    # as desired
    actions: Mapped[str] = mapped_column(String, nullable=False)
    # TODO: ditto
    legendary_actions: Mapped[str] = mapped_column(String, nullable=True)
    # TODO: ditto
    special_abilities: Mapped[str] = mapped_column(String)

    senses: Mapped[str]

    damage_resistances: Mapped[str] = mapped_column(String, nullable=True)
    damage_immunities: Mapped[str] = mapped_column(String, nullable=True)
    condition_immunities: Mapped[str] = mapped_column(String, nullable=True)
    speeds: Mapped[str] = mapped_column(String, nullable=False)
    saving_throws: Mapped[str] = mapped_column(String)
    # END NOTE
    
    languages: Mapped["MonsterLanguage"] = relationship("MonsterLanguage", back_populates="monster")
    spells: Mapped["MonsterSpell"] = relationship("MonsterSpell", back_populates="monster")



    def __init__(self, *args, **kwargs):
        super().__init__()
        self.calculate_and_store_ability_scores({
            'strength': args.strength,
            'dexterity': args.dexterity,
            'constitution': args.constitution,
            'intelligence': args.intelligence,
            'wisdom': args.wisdom,
            'charisma': args.charisma
        })


    def calculate_and_store_ability_scores(self, ability_scores: Dict[str, int]):
        for ability_name, score in ability_scores:
            # 1. dynamically assign attribute by string literal
            setattr(self, ability_name, score)
            # 2. call calculate_ability_modifier
            modifier = Monster.calculate_ability_modifier()
            # 3. store that dynamically as well - assume column naming schema is unchanged
            setattr(self, f"{ability_name}_mod", modifier)


    @staticmethod
    def calculate_ability_modifier(ability_score:int):
        return (ability_score - 10) // 2 

class Armor(Base):
    __tablename__ = "armors"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)    
    name: Mapped[str] = mapped_column(String(100), primary_key=True)
    weight: Mapped[int] = mapped_column(Integer, nullable=False)
    description: Mapped[str] = mapped_column(String, nullable=True)
    strength_requirement: Mapped[int] = mapped_column(Integer, default=0)
    armor_class: Mapped[int] = mapped_column(Integer, nullable=False)
    max_dex_bonus: Mapped[int] = mapped_column(Integer, nullable=False)
    stealth_disadvantage: Mapped[Boolean] = mapped_column(Boolean, nullable=False)
    ac_enchantment: Mapped[int] = mapped_column(Integer, default=0)
    special_effects: Mapped[str] = mapped_column(String)
    requires_attunement: Mapped[Boolean] = mapped_column(Boolean, nullable=False)


class MonsterArmor(Base):
    """
    Junction table for Monster *-* Armor
    """
    __tablename__ = 'monster_armor'
    
    monster_id: Mapped[int] = mapped_column(Integer, ForeignKey('monsters.id'), primary_key=True, unique=True)
    armor_id: Mapped[int] = mapped_column(Integer, ForeignKey('armors.id'), primary_key=True, unique=True)
    
    # Relationship to Monster and Armor
    monster: Mapped["Monster"] = relationship("Monster", back_populates="armors")
    armor: Mapped["Armor"] = relationship("Armor")


class CreatureType(Base):
    """
    Creature type

    Example: "humanoid"
    """
    __tablename__ = "creature_types"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(50), nullable=False, primary_key=True)    

class CreatureSubType(Base):
    """
    Creature sub-type

    Example: "goblinoid"
    """
    __tablename__ = "creature_sub_types"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(50), nullable=False, primary_key=True)    

class CreatureSize(Base):
    """
    Creature size, as defined in DND5e(2014) PHB p.191
    """
    __tablename__ = "creature_sizes"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(20), nullable=False, primary_key=True)
    # NOTE: what about Gargantuan? It is a range?
    space = mapped_column(Float, nullable=False)
    


class ActionAttack(Base):
    __tablename__ = "actions"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(100), nullable=False, primary_key=True)


class Language(Base):
    __tablename__ = "languages"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(50), nullable=False, primary_key=True)


class Spell(Base):
    __tablename__ = "spells"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(300), nullable=False, primary_key=True)
    
    # NOTE: we handle "at higher levels", by adding a new entry with a different level
    level = mapped_column(Integer, primary_key=True)
    magic_school_id: Mapped[int] = mapped_column(Integer, ForeignKey('magic_schools.id'))
    magic_school: Mapped["MagicSchool"] = relationship(
        "MagicSchool",
    )
    
    classes: Mapped["SpellClass"] = relationship("SpellClass", back_populates="spell")

    # NOTE: this should be computed and converted to combat rounds, hours, 
    # days, etc, as appropriate, per context
    duration_minutes: Mapped[int] = mapped_column(Integer)
    casting_time_id: Mapped[str] = mapped_column(String, nullable=False)
    range: Mapped[int] = mapped_column(Integer)
    area_of_effect_id: Mapped[int] = mapped_column(Integer, ForeignKey('area_of_effects.id'))
    area_of_effect: Mapped["AreaOfEffect"] = relationship(
        "AreaOfEffect",
    )
    concentration: Mapped[bool] = mapped_column(Boolean)

    components: Mapped[str] = mapped_column(String)


class CharacterClass(Base):
    """
    TODO: finish this up; as minimal columns set as possible
    """
    __tablename__ = "character_classes"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)


class SpellClass(Base):
    """
    Junction Table for Spell *-* CharacterClass
    """
    __tablename__ = "spell_to_class"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)

    class_id: Mapped[int] = mapped_column(Integer, ForeignKey('character_classes.id'), primary_key=True)
    character_class: Mapped["CharacterClass"] = relationship("CharacterClass")
    
    spell_id: Mapped[int] = mapped_column(Integer, ForeignKey('spells.id'), primary_key=True)
    spell: Mapped["Spell"] = relationship("Spell", back_populates="classes")


class MonsterSpell(Base):
    """
    Junciton table for spell *-* mosnter
    """
    __tablename__ = "spell_to_monster"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)

    monster_id: Mapped[int] = mapped_column(Integer, ForeignKey('monsters.id'),primary_key=True, unique=True)
    monster: Mapped["Monster"] = relationship('Monster', back_populates="spells")

    spell_id: Mapped[int] = mapped_column(Integer, ForeignKey('spells.id'), primary_key=True, unique=True)
    spell: Mapped["Spell"] = relationship("Spell")


class MonsterLanguage(Base):
    """
    Junction table for Language *-* Monster
    """
    __tablename__ = "monster_language"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)

    monster_id: Mapped[int] = mapped_column(Integer, ForeignKey('monsters.id'), primary_key=True, unique=True)
    monster: Mapped["Monster"] = relationship('Monster')

    language_id: Mapped[int] = mapped_column(Integer, ForeignKey('languages.id'), primary_key=True, unique=True)
    language: Mapped["Language"] = relationship('Language')



class AreaOfEffect(Base):
    __tablename__ = "area_of_effects"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(20), nullable=False, primary_key=True)
    desc: Mapped[str] = mapped_column(String, nullable=False)


class MagicSchool(Base):
    __tablename__ = "magic_schools"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, unique=True)
    name: Mapped[str] = mapped_column(String(20), nullable=False, primary_key=True)
    desc: Mapped[str] = mapped_column(String, nullable=False)
    




"""
Util functions
"""
from random import randrange
from typing import Dict, List, Tuple


def get_random_elem_from_json_table(table: List[Dict]) -> Dict:
    """
    Gets an element from a dict-serialized table
    """
    range_upper = len(table)
    random_index = randrange(0, range_upper)
    return table[random_index]


# TODO: this is basically pseudocode. Need to finalize based on actual dict
# TODO: implement as binary search, just for optimality's sake
def get_elem_in_range_from_json_table(table: List[Dict], index: int) -> Tuple[Dict, int]:
    """
    Gets an element from a dict-serialized table, along with the index
    of its location in the table
    """
    for i in range(len(table)):
        element = table[i]
        range_values = element["range"]
        # because python ranges are non-inclusive and json ranges are inclusive
        range_values[-1] += 1
        lower_range, upper_range = range_values
        # NOTE: may need to adjust
        if index in range(lower_range, upper_range):
            return element, i


def roll(dice_max: int, modifiers: list[int] = None) -> int:
    """

    Handle a dice roll
    
    Args:
        dice_max (int): _description_
        modifiers (list, optional): _description_. Defaults to [].

    Returns:
        int: _description_
    """
    range_upper = dice_max+1
    roll_result = randrange(0, range_upper)
    for modifier in modifiers:
        roll_result += modifier
    return roll_result


def roll_with_advantage(dice_max: int, modifiers: list[int] = None) -> int:
    roll_1 = roll(dice_max, modifiers)
    roll_2 = roll(dice_max, modifiers)
    return max(roll_1, roll_2)


def roll_with_disadvantage(dice_max: int, modifiers: list[int] = None) -> int:
    roll_1 = roll(dice_max, modifiers)
    roll_2 = roll(dice_max, modifiers)
    return min(roll_1, roll_2)



import os
import importlib
from json import load
from typing import Dict, Any, List


def list_json_files_in_subdir(package: str) -> List[str]:
    """
    List all JSON files in the specified subdirectory package.

    Args:
        package (str): The name of the package to list JSON files from.

    Returns:
        List[str]: List of filenames ending in .json
    """
    try:
        # Dynamically load the specified package using importlib
        package_module = importlib.import_module(package)
        package_path = os.path.dirname(package_module.__file__)

        # List all files in the specified package and filter for JSON files
        json_files = [
            file for file in os.listdir(package_path) if file.endswith(".json")
        ]
        return json_files

    except Exception as e:
        raise Exception(f"Failed to list JSON files: {e}")


def prepare_json_table(path_to_table: str) -> list[Dict]:
    """
    Deserialized a json file representing a table into Python types,
    for subsequent operations

    Args:
        table (json): json file

    Returns:
        list[Dict]
    """
    reserialized_object = None
    with open(path_to_table, 'r', encoding='utf-8') as file:
        reserialized_object = load(file)

    return reserialized_object


def get_json_table(path_to_table: str) -> list[Dict]:
    """

    Args:
        path_to_table (str): _description_
    """
    print(f"path_to_table: {path_to_table}")



import os
from dotenv import load_dotenv
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

load_dotenv()
DB_USER = os.getenv("DB_USER", 'default_user')
DB_PASSWORD = os.getenv("DB_PASSWORD", 'default_password')
DB_HOST = os.getenv("DB_HOST", "default_host")
DB_PORT = os.getenv("DB_PORT", "default_port")
DB_NAME = os.getenv("DB_NAME", "default_name")
DB_ENGINE = os.getenv("DB_ENGINE", "postgresql")

DB_URI = f'{DB_ENGINE}://{DB_USER}:{DB_PASSWORD}@{DB_HOST}:{DB_PORT}/{DB_NAME}'

engine = create_engine(DB_URI)

# all models should inherit this
Base = declarative_base()

# handles db sessions
Session = sessionmaker(autocommit=False, autoflush=False, bind=engine)

def get_db():
    db = Session()

    try:
        yield db
    except:
        Exception("error getting db session")
    
    finally:
        db.close()



"""
Module used for all CR operations, based on logic defined in 
DND 5e(2014) DMG p.81-85
"""
import os
from math import inf
from typing import List, Dict, Tuple
from utils.utils import list_json_files_in_subdir, prepare_json_table
from dm_utils.utils import get_elem_in_range_from_json_table

TABLES_SUBDIR = "dm_tables_json.encounter_planning"
TABLE_FILENAMES = list_json_files_in_subdir(
    TABLES_SUBDIR
)

CWD = os.getcwd()
TABLE_FILEPATHS = dict()
for filename in TABLE_FILENAMES:
    file_key = filename.split('.')[0]
    full_path_to_file = os.path.join(
        CWD,
        TABLES_SUBDIR.replace('.', '/'),
        filename
    )
    TABLE_FILEPATHS[file_key] = full_path_to_file
    
XP_THRESHOLDS_TABLE = prepare_json_table(
    TABLE_FILEPATHS['xp_thresholds_by_character_level']
)
ENCOUNTER_MULTIPLIERS_TABLE = prepare_json_table(
    TABLE_FILEPATHS['encounter_multipliers']
)
ADVENTURING_DAY_XP_TABLE = prepare_json_table(
    TABLE_FILEPATHS['adventuring_day_xp_by_level']
)


#TODO: add some dataclasses, don't be a pussy
def determine_party_xp_thresholds(party_levels: List[int]) -> List[int]:
    threshold_values = {
        "easy": {
            "sum": 0,
            "xp_thresholds": []
        },
        "medium": {
            "sum": 0,
            "xp_thresholds": []
        },
        "hard": {
            "sum": 0,
            "xp_thresholds": []
        },
        "deadly": {
            "sum": 0,
            "xp_thresholds": []
        },
    }

    for level in party_levels:
        thresholds_level_row = XP_THRESHOLDS_TABLE[str(level)]

        easy_threshold = thresholds_level_row["easy"]
        threshold_values["easy"]["xp_thresholds"].append(easy_threshold)
        threshold_values["easy"]["sum"] += easy_threshold
        
        medium_threshold = thresholds_level_row["medium"]
        threshold_values["medium"]["xp_thresholds"].append(medium_threshold)
        threshold_values["medium"]["sum"] += medium_threshold
        
        hard_threshold = thresholds_level_row["hard"]
        threshold_values["hard"]["xp_thresholds"].append(hard_threshold)
        threshold_values["hard"]["sum"] += hard_threshold
        
        deadly_threshold = thresholds_level_row["deadly"]
        threshold_values["deadly"]["xp_thresholds"].append(deadly_threshold)
        threshold_values["deadly"]["sum"] += deadly_threshold
        
    return threshold_values
        
    
# TODO: define interface for monsters
def get_total_monster_xp(monsters: List[Dict]) -> int:
    sum_monster_xp = 0
    # assume someting like: {name: {..., xp:int,...},...}
    for monster in monsters:
        sum_monster_xp += monster['xp']
    
    return sum_monster_xp


def adjust_xp_multiplier_by_party_size(
    current_multiplier_table_index: float,
    higher_xp_threshold: bool = True
) -> int:
    if higher_xp_threshold:
        closest_multiplier_row = \
            ENCOUNTER_MULTIPLIERS_TABLE[
                current_multiplier_table_index + 1
            ]
    else:
        closest_multiplier_row = \
            ENCOUNTER_MULTIPLIERS_TABLE[
                current_multiplier_table_index - 1
            ]
    return closest_multiplier_row["multiplier"]

    
def get_encounter_multiplier(party_size: int, num_monsters: int) -> float:
    encounter_multiplier_row, table_index = get_elem_in_range_from_json_table(
        ENCOUNTER_MULTIPLIERS_TABLE, 
        num_monsters
    )
    encounter_multiplier = encounter_multiplier_row['multiplier']
    # adjust, according to party size
    # DMG defines default in range of 3-5 party members
    default_party_size_range = range(3, 6)
    
    if party_size not in default_party_size_range:
        # interesting syntax
        if party_size < default_party_size_range[0]:
            adjust_xp_multiplier_by_party_size(
                table_index,
                higher_xp_threshold=True
            )
        elif num_monsters == 1:
            return 0.5
        else:
            adjust_xp_multiplier_by_party_size(
                table_index,
                higher_xp_threshold=False
            )
    
    return encounter_multiplier
    
    
def get_closest_difficulty_threshold(
    xp_thresholds: Dict,
    monster_xp: int
) -> Tuple[str, int]:
    """
    {
        "closest_difficulty": str[easy, medium, hard],
        "monster_xp":
    }
    """
    closest_difficulty = ""
    closest_difference = inf

    for difficulty, threshold_values in xp_thresholds.items():
        xp_thresholds_sum = threshold_values["sum"]
        difference = abs(monster_xp - xp_thresholds_sum)
        
        if difference < closest_difference:
            closest_difference = difference
            closest_difficulty = difficulty

    return closest_difficulty, closest_difference

    
def get_daily_xp_limit_per_character_level(
    character_level: int
) -> int:
    return ADVENTURING_DAY_XP_TABLE[character_level]


def get_adventuring_day_xp_thresholds(
    party_levels: List[int]
) -> int:
    """
    Assuming the party will take at most two short rests between encounters

    Args:
        party_levels (List[int]): list of int values, representing character 
        levels in the party

    Returns:
        int: the upper bound of total xp that is recommended the party to
        receive from defeating monsters
    """
    party_xp_sum = 0
    for character_level in party_levels:
        character_xp_limit = \
            get_daily_xp_limit_per_character_level(character_level)
        party_xp_sum += character_xp_limit
    return party_xp_sum
    

def get_encounter_difficulty_data(
    party_levels: List[int],
    encounter_monsters: List[Dict]
) -> Dict:
    """
    Main driver function to calculate encounter CR
    Based on logic in 5e(2014) DMG p.82
    """
    # TODO: check the calling scope. May make more sense to 
    # already have the xp_thresolds, total_monster_xp calculated 
    # already and just inject them
    xp_thresholds: Dict = determine_party_xp_thresholds(party_levels)
    total_monster_xp = get_total_monster_xp(encounter_monsters)        
    encounter_multiplier = get_encounter_multiplier(
        len(party_levels),
        len(encounter_monsters)
    )
    total_monster_xp *= encounter_multiplier
    
    # place total_monster_xp in xp_thresholds
    closest_difficulty = get_closest_difficulty_threshold(
        xp_thresholds,
        total_monster_xp
    )
    
    # TODO: this shit needs type definiton via dataclasses
    encounter_calculation_data = {
        "encounter_difficulty": closest_difficulty,
        "total_monster_xp": total_monster_xp,
        "encounter_multiplier": encounter_multiplier,
        "xp_thresholds": xp_thresholds
    }
    
    return encounter_calculation_data


# TODO: move this to tests later
if __name__ == "__main__":
    # example from p.83
    character_levels = [3, 3, 3, 2]
    monsters = [
        {
            "name": "Bugbear",
            "xp": 200
        },
        {
            "name": "Hobgoblin",
            "xp": 100
        },
        {
            "name": "Hobgoblin",
            "xp": 100
        },
        {
            "name": "Hobgoblin",
            "xp": 100
        },
    ]
    # expect encounter_difficulty to be "hard"
    calculated_difficulty = get_encounter_difficulty_data(
        character_levels,
        monsters
    )
    print(f'DEBUG: {calculated_difficulty}')


from random import randrange


def hit_check(attack_roll:int, armor_class:int)->bool:
    return attack_roll >= armor_class


def calculate_initiative(initative_mod: int) -> int:
    dice_roll = randrange(1, 21)
    return dice_roll + initative_mod

# TODO: formula should be codified as a typesomehow, somewhere

def calculate_hp(hit_point_formula: str) -> int:
    """_summary_

    Args:
        hit_point_formula (str): follows this format:
        2d8+2©

    Returns:
        int: _description_
    """
    hp = 0
    formula_parts = hit_point_formula.split("+")
    con_mod_hp_bonus = int(formula_parts[-1])
    
    num_rolls, dice_size = formula_parts[0].split('d')
    num_rolls = int(num_rolls)
    dice_size = int(dice_size)
    
    for i in range(num_rolls):
        dice_roll = randrange(1, dice_size+1)
        hp += dice_roll
    
    dice_roll += con_mod_hp_bonus
    return dice_roll


# TODO: move to tests
if __name__ == "__main__":
    dice_formula = "2d8 + 2"
    calculated_hp = calculate_hp(dice_formula)
    print(f'calculated_hp: {calculated_hp}')
